from frasco.mail.provider import MailProvider
from frasco.utils import inject_app_config


class SMTPProvider(MailProvider):
    def send(self, msg):
        return self.state.mail.send(msg)

    def start_bulk_connection(self):
        connection = self.state.mail.connect()
        # simulate entering a with context
        # (flask-mail does not provide a way to connect otherwise)
        connection.__enter__()
        return connection

    def stop_bulk_connection(self, conn):
        conn.__exit__(None, None, None)



from flask import Flask
from flask.ctx import has_request_context
from frasco.config import Config
from frasco.utils import JSONEncoder, AttrDict
from frasco.templating import configure_environment
from werkzeug.contrib.fixers import ProxyFix
import os
import sys


class Frasco(Flask):
    config_class = Config
    json_encoder = JSONEncoder

    def __init__(self, import_name, **kwargs):
        super(Frasco, self).__init__(import_name, **kwargs)
        self.extensions = AttrDict()
        self.wsgi_app = ProxyFix(self.wsgi_app)

    def log_exception(self, exc_info):
        if not isinstance(exc_info, tuple):
            exc_info = sys.exc_info()
        if has_request_context():
            super(Frasco, self).log_exception(exc_info)
        else:
            self.logger.error('Exception [outside of request context]', exc_info=exc_info)

    def create_jinja_environment(self):
        env = super(Frasco, self).create_jinja_environment()
        env.globals.update(app=self)
        configure_environment(env)
        macro_file = os.path.join(self.root_path, "macros.html")
        if os.path.exists(macro_file):
            env.macros.register_file(macro_file)
        macro_dir = os.path.join(self.root_path, "macros")
        if os.path.exists(macro_dir):
            env.macros.register_directory(macro_dir)
        return env



from frasco.models import db, transaction
from frasco.models.utils import MutableList
from frasco.redis import redis
from flask_login import UserMixin
from sqlalchemy.dialects import postgresql
import datetime


__all__ = ('UserModelMixin', 'UserWithUsernameModelMixin', 'UserLastAccessAtModelMixin', 'UserLoginModelMixin')


class UserModelMixin(UserMixin):
    email = db.Column(db.String)
    password = db.Column(db.String, nullable=True)
    last_password_change_at = db.Column(db.DateTime)
    last_password_change_from = db.Column(db.String)
    previous_passwords = db.Column(MutableList.as_mutable(postgresql.ARRAY(db.String)))
    must_reset_password_at_login = db.Column(db.Boolean, default=False)
    signup_at = db.Column(db.DateTime, default=datetime.datetime.utcnow)
    signup_from = db.Column(db.String)
    signup_provider = db.Column(db.String)
    signup_country = db.Column(db.String)
    last_login_at = db.Column(db.DateTime)
    last_login_from = db.Column(db.String)
    last_login_provider = db.Column(db.String)
    updated_at = db.Column(db.DateTime)
    auth_providers = db.Column(MutableList.as_mutable(postgresql.ARRAY(db.String)), default=list)

    @classmethod
    def query_by_email(cls, email):
        return cls.query.filter(cls.email == email.lower())

    @classmethod
    def query_by_identifier(cls, identifier):
        return cls.query_by_email(identifier)


class UserWithUsernameModelMixin(UserModelMixin):
    username = db.Column(db.String, unique=True)

    @classmethod
    def query_by_username(cls, username):
        return cls.query.filter(db.func.lower(cls.username) == username.strip().lower())

    @classmethod
    def query_by_username_or_email(cls, identifier):
        return cls.query.filter(db.or_(db.func.lower(cls.username) == identifier.strip().lower(), cls.email == identifier.strip().lower()))

    @classmethod
    def query_by_identifier(cls, identifier):
        return cls.query_by_username(identifier)


class UserLastAccessAtModelMixin(object):
    last_access_at = db.Column(db.Date)

    def update_last_access_at(self):
        if not self.id:
            return
        today_key = "users-last-access-%s" % datetime.date.today().isoformat()
        if not redis.getbit(today_key, self.id):
            with transaction():
                self.last_access_at = datetime.date.today()
            redis.setbit(today_key, self.id, 1)
            redis.expire(today_key, 86500)


class UserLoginModelMixin(object):
    login_at = db.Column(db.DateTime)
    login_from = db.Column(db.String)
    login_provider = db.Column(db.String)
    login_country = db.Column(db.String)
    login_user_agent = db.Column(db.String)



from flask import flash, has_request_context, request
from frasco.ext import get_extension_state
from flask_bcrypt import Bcrypt
import datetime
import re
import logging

from .tokens import generate_user_token, read_user_token
from .signals import reset_password_sent, password_updated


__all__ = ('hash_password', 'PasswordValidationFailedError', 'validate_password', 'update_password',
            'check_password', 'generate_reset_password_token', 'send_reset_password_token')


bcrypt = Bcrypt()
logger = logging.getLogger('frasco.users')


def hash_password(password):
    return bcrypt.generate_password_hash(password)


def check_password(user, password):
    return user.password and bcrypt.check_password_hash(user.password, password)


class PasswordValidationFailedError(Exception):
    def __init__(self, reason=None, rule=None):
        super(PasswordValidationFailedError, self).__init__()
        self.reason = reason
        self.rule = rule

    def __unicode__(self):
        return self.reason


def validate_password(user, password, flash_messages=True, raise_error=True, pwhash=None):
    state = get_extension_state('frasco_users')
    pwhash = hash_password(password) if not pwhash else pwhash

    if state.options['min_time_between_password_change'] and user.last_password_change_at and not user.must_reset_password_at_login:
        if (datetime.datetime.utcnow() - user.last_password_change_at).total_seconds() < state.options['min_time_between_password_change']:
            if flash_messages and state.options['min_time_between_password_change_message']:
                flash(state.options['min_time_between_password_change_message'], 'error')
            if raise_error:
                raise PasswordValidationFailedError("password_change_too_soon")
            return False

    if state.options['validate_password_regexps']:
        for pattern, label in state.options['validate_password_regexps']:
            if not re.search(pattern, password):
                if flash_messages and state.options['validate_password_regexps_message']:
                    flash(state.options['validate_password_regexps_message'].format(rule=label), 'error')
                if raise_error:
                    raise PasswordValidationFailedError("invalid_password", label)
                return False

    if state.options['prevent_password_reuse'] and user.password:
        for oldhash in [user.password] + (user.previous_passwords or []):
            if oldhash and bcrypt.check_password_hash(oldhash, password):
                if flash_messages and state.options['password_reused_message']:
                    flash(state.options['password_reused_message'], 'error')
                if raise_error:
                    raise PasswordValidationFailedError("password_reused")
                return False

    for validator in state.password_validators:
        if not validator(password):
            if raise_error:
                raise PasswordValidationFailedError()
            return False

    return True


def update_password(user, password, skip_validation=False, **kwargs):
    """Updates the password of a user
    """
    state = get_extension_state('frasco_users')
    pwhash = hash_password(password)
    if not skip_validation and not validate_password(user, password, pwhash=pwhash, **kwargs):
        return False
    if state.options['prevent_password_reuse']:
        user.previous_passwords = [user.password] + (user.previous_passwords or [])
        if state.options['max_password_reuse_saved']:
            user.previous_passwords = user.previous_passwords[:state.options['max_password_reuse_saved']]
    user.password = pwhash
    user.last_password_change_at = datetime.datetime.utcnow()
    if has_request_context():
        user.last_password_change_from = request.remote_addr
    user.must_reset_password_at_login = False
    password_updated.send(user=user)
    logger.info('Password changed for user #%s' % user.id)
    return True


def generate_reset_password_token(user):
    return generate_user_token(user, salt="password-reset")


def send_reset_password_token(user):
    """Generates a reset password token and optionnaly (default to yes) send the reset
    password email
    """
    from frasco.mail import send_mail
    token = generate_reset_password_token(user)
    send_mail(user.email, "users/reset_password.txt", user=user, token=token, locale=getattr(user, 'locale', None))
    reset_password_sent.send(user=user, token=token)
    logger.info('Password reset token sent to user #%s' % user.id)
    return token



from frasco.upload.backend import StorageBackend
from frasco.helpers import url_for
from flask import safe_join, current_app
import os


class LocalStorageBackend(StorageBackend):
    def save(self, file, filename):
        filename = safe_join(os.path.abspath(self.options["upload_dir"]), filename)
        dirname = os.path.dirname(filename)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        file.save(filename)

    def url_for(self, filename, **kwargs):
        return url_for("static_upload", filename=filename, **kwargs)

    def delete(self, filename):
        filename = safe_join(os.path.abspath(self.options["upload_dir"]), filename)
        if os.path.exists(filename):
            os.unlink(filename)



from flask.signals import Namespace


_signals = Namespace()
translation_extracted = _signals.signal("translation_extracted")
translation_updated = _signals.signal("translation_updated")
translation_compiled = _signals.signal("translation_compiled")



from flask import current_app
from flask.cli import AppGroup
from frasco.ext import get_extension_state
from frasco.utils import shell_exec
import click
import tempfile
import os
import re

from .extract import edit_pofile, create_babel_mapping, exec_babel_extract, po_to_json
from .signals import *


babel_cli = AppGroup('babel', help='Commands to manage translations')


def _extract():
    state = get_extension_state('frasco_babel')
    path = os.path.join(current_app.root_path, "translations")
    if not os.path.exists(path):
        os.mkdir(path)
    potfile = os.path.join(path, "messages.pot")

    click.echo("Extracting translatable strings from %s" % path)
    mapping = create_babel_mapping(state.options["extract_jinja_dirs"],
        state.options["extract_with_jinja_exts"], state.options["extractors"])
    exec_babel_extract(current_app.root_path, potfile, mapping)

    # we need to extract message from other paths independently then
    # merge the catalogs because babel's mapping configuration does
    # not support absolute paths
    for path, jinja_dirs, jinja_exts, extractors in state.extract_dirs:
        click.echo("Extracting translatable strings from %s" % path)
        mapping = create_babel_mapping(jinja_dirs, jinja_exts, extractors)
        path_potfile = tempfile.NamedTemporaryFile(delete=False)
        path_potfile.close()
        exec_babel_extract(potfile, path_potfile.name, mapping)
        os.unlink(path_potfile.name)

    translation_extracted.send()


@babel_cli.command('extract')
def extract_cmd():
    _extract()


@babel_cli.command("init")
@click.argument('locale')
@click.option('--gotrans/--no-gotrans', default=False)
def init_translation(locale, gotrans=False):
    state = get_extension_state('frasco_babel')
    path = os.path.join(current_app.root_path, "translations")
    potfile = os.path.join(path, "messages.pot")
    if not os.path.exists(potfile):
        _extract()
    click.echo("Initializing new translation '%s' in %s" % (locale, os.path.join(path, locale)))
    shell_exec([state.options["babel_bin"], "init", "-i", potfile, "-d", path, "-l", locale])
    translation_updated.send(None, locale=locale)
    if gotrans:
        translate_with_google(locale)


@babel_cli.command("update")
@click.option('--extract/--no-extract', default=True)
@click.option('--gotrans/--no-gotrans', default=False)
def update_translations(extract=True, gotrans=False):
    state = get_extension_state('frasco_babel')
    path = os.path.join(current_app.root_path, "translations")
    potfile = os.path.join(path, "messages.pot")
    if not os.path.exists(potfile) or extract:
        _extract()
    click.echo("Updating all translations")
    shell_exec([state.options["babel_bin"], "update", "-i", potfile, "-d", path])
    for f in os.listdir(path):
        if os.path.isdir(os.path.join(path, f)):
            translation_updated.send(locale=f)
            if gotrans:
                translate_with_google(f)


@babel_cli.command("compile")
def compile_translations():
    click.echo("Compiling all translations")
    state = get_extension_state('frasco_babel')
    path = os.path.join(current_app.root_path, "translations")
    shell_exec([state.options["babel_bin"], "compile", "-d", path])
    if state.options['compile_to_json']:
        output = os.path.join(current_app.static_folder, state.options['compile_to_json'])
        for f in os.listdir(path):
            if os.path.isdir(os.path.join(path, f)):
                _po2json(f, output % f)
    if state.options['compile_to_js']:
        output = os.path.join(current_app.static_folder, state.options['compile_to_js'])
        for f in os.listdir(path):
            if os.path.isdir(os.path.join(path, f)):
                _po2js(f, output % f)
    translation_compiled.send()


def _po2json(locale, output=None):
    filename = os.path.join(current_app.root_path, "translations", locale, "LC_MESSAGES", "messages.po")
    dump = po_to_json(filename)
    if output:
        with open(output, 'w') as f:
            f.write(dump)
    else:
        click.echo(dump)


def _po2js(locale, output=None):
    filename = os.path.join(current_app.root_path, "translations", locale, "LC_MESSAGES", "messages.po")
    varname = get_extension_state('frasco_babel').options['js_catalog_varname']
    dump = "var %s = %s;" % (varname % locale.upper(), po_to_json(filename))
    if output:
        with open(output, 'w') as f:
            f.write(dump)
    else:
        click.echo(dump)


@babel_cli.command('po2json')
@click.argument('locale')
@click.option('--output', '-o')
def po2json(locale, output=None):
    _po2json(locale, output)


@babel_cli.command('po2js')
@click.argument('locale')
@click.option('--output', '-o')
def po2js(locale, output=None):
    _po2js(locale, output)


@babel_cli.command("gotrans")
@click.argument('locale')
def translate_with_google(locale):
    import goslate
    click.echo("Google translating '%s'" % locale)
    click.echo("WARNING: you must go through the translation after the process as placeholders may have been modified", fg="red")
    filename = os.path.join(current_app.root_path, "translations", locale, "LC_MESSAGES", "messages.po")
    
    def translate(id):
        # google translate messes with the format placeholders thus
        # we replace them with something which is easily recoverable
        string, placeholders = safe_placeholders(id)
        string = gs.translate(string, locale)
        return unsafe_placeholders(string, placeholders, "## %s ##")

    with edit_pofile(filename) as catalog:
        gs = goslate.Goslate()
        for message in catalog:
            if not message.id:
                continue
            if message.pluralizable:
                string = list(message.string)
                if not string[0]:
                    string[0] = translate(message.id[0])
                if not string[1]:
                    string[1] = translate(message.id[1])
                message.string = tuple(string)
            elif not message.string:
                message.string = translate(message.id)


def safe_placeholders(string, repl="##%s##"):
    placeholders = []
    def replace_placeholder(m):
        placeholders.append(m.group(1))
        return repl % (len(placeholders) - 1)
    string = re.sub(r"%\(([a-zA-Z_]+)\)s", replace_placeholder, string)
    return string, placeholders


def unsafe_placeholders(string, placeholders, repl="##%s##"):
    for i, placeholder in enumerate(placeholders):
        string = string.replace(repl % i, "%%(%s)s" % placeholder)
    return string
